import 'dart:async';
import 'dart:io';
import 'package:quiver/pattern.dart';

/// A [Handler] can handle [HttpRequest] specific operations
/// or act as middleware.
typedef Handler = FutureOr<void> Function(HttpRequest, [Context]);

/// Context for the current route pipeline.
/// The [Context] gets passed from middleware to middleware.
/// Anything that gets passed to [values] will be visibel to the other middleware.
/// Calling [cancel] prevents any further middlewere from executing.
class Context {
  var _canceled = false;
  bool get canceled => _canceled;
  final values = <String, dynamic>{};
  Context();
  void cancel() {
    _canceled = true;
  }
}

/// Executes middleware before and after [_handlerFunc].
/// All middleware gets executed sequentially.
class MiddlewareHandler {
  static const _emptyMiddleware = <Handler>[];
  final Handler _handlerFunc;
  final List<Handler> _before, _after;

  /// [_handlerFunc] is the operation specific [Handler].
  /// [before] is all the middleware that gets executed before [_handlerFunc].
  /// [after] is all the middleware that gets executed before [_handlerFunc].
  MiddlewareHandler(this._handlerFunc,
      {List<Handler> before = _emptyMiddleware,
      List<Handler> after = _emptyMiddleware})
      : _before = before,
        _after = after;

  FutureOr<void> call(HttpRequest request, [Context context]) async {
    context ??= Context();
    for (final middleware in _before) {
      if (context.canceled) return;
      await middleware(request, context);
    }

    if (context.canceled) return;
    await _handlerFunc(request, context);

    for (final middleware in _after) {
      if (context.canceled) return;
      await middleware(request, context);
    }
  }
}

/// It could get tedious creating multiple [MiddlewareHandler]s with the same middleware.
/// [MiddlewareHandlerBuilder] makes it simpler.
class MiddlewareHandlerBuilder {
  final List<Handler> _before, _after;

  MiddlewareHandlerBuilder(
      {List<Handler> before = MiddlewareHandler._emptyMiddleware,
      List<Handler> after = MiddlewareHandler._emptyMiddleware})
      : _before = before,
        _after = after;

  MiddlewareHandler build(Handler handlerFunc) {
    return MiddlewareHandler(handlerFunc, before: _before, after: _after);
  }
}

/// Handles all the Http method specific [Handler]s and executes the right one.
class HttpMethodHandler {
  static FutureOr<void> _standardBadRequestHandler(HttpRequest request,
      [Context context]) {
    request.response.statusCode = HttpStatus.badRequest;
    //ignore:unawaited_futures
    request.response.close();
    context.cancel();
  }

  final Handler _get,
      _post,
      _delete,
      _update,
      _push,
      _connect,
      _options,
      _trace,
      _patch,
      _badRequest;

  /// [badRequest] gets executed when either the Http method is bogus
  /// or when no handler for that method has been registered.
  HttpMethodHandler(
      {Handler get,
      Handler delete,
      Handler update,
      Handler post,
      Handler push,
      Handler connect,
      Handler options,
      Handler trace,
      Handler patch,
      Handler badRequest = _standardBadRequestHandler})
      : _get = get,
        _post = post,
        _delete = delete,
        _update = update,
        _push = push,
        _connect = connect,
        _options = options,
        _trace = trace,
        _patch = patch,
        _badRequest = badRequest;

  FutureOr<void> call(HttpRequest request, [Context context]) async {
    context ??= Context();
    Handler handler;
    switch (request.method) {
      case 'GET':
        handler = _get;
        break;
      case 'POST':
        handler = _post;
        break;
      case 'UPDATE':
        handler = _update;
        break;
      case 'DELETE':
        handler = _delete;
        break;
      case 'PUSH':
        handler = _push;
        break;
      case 'CONNECT':
        handler = _connect;
        break;
      case 'OPTIONS':
        handler = _options;
        break;
      case 'TRACE':
        handler = _trace;
        break;
      case 'PATCH':
        handler = _patch;
        break;
      default:
        _badRequest(request, context);
        return;
    }
    if (handler == null) {
      _badRequest(request, context);
      return;
    }
    await handler(request, context);
  }
}

class Route {
  final String route;
  final Glob routeGlob;
  final HttpMethodHandler _methodHandler;

  Route(this.route, this._methodHandler) : routeGlob = Glob(route);
}

/// Matches incoming [HttpRequests] against registered [Routes].
///
/// Caution: [Routes] get matched in the order they are registered.
/// Overshadowing can be detected by enabling assertions.
class Router {
  static FutureOr<void> _standardNotFoundHandler(HttpRequest req,
      [Context context]) {
    req.response.statusCode = HttpStatus.notFound;
    req.response.writeln(req.response.reasonPhrase);
    req.response.close();
  }

  final List<Route> _routes = [];
  final Handler _notFound;

  /// [notFound] gets executed when no [Route] matches the [HttpRequest]
  Router({Handler notFound = _standardNotFoundHandler}) : _notFound = notFound;

  /// Checks if [route] would be overshadowed by an already registered [Route]
  bool routeOvershadowed(Route route) {
    final index = _routes.indexWhere((r) => r.routeGlob.hasMatch(route.route));
    return index != -1;
  }

  void register(Route route) {
    assert(
        !routeOvershadowed(route), 'Route "${route.route} gets overshadowed"');

    _routes.add(route);
  }

  /// Handle incoming [HttpRequest]
  FutureOr<void> handle(HttpRequest request, [Context context]) async {
    context ??= Context();
    final index =
        _routes.indexWhere((r) => r.routeGlob.hasMatch(request.uri.path));
    if (index == -1) {
      _notFound(request, context);
      return;
    }
    final route = _routes[index];
    await route._methodHandler(request, context);
  }
}
